#!/usr/bin/perl

# Copyright 2013, 2015 Kevin Ryde.

# This file is part of miscbits-el.
#
# miscbits-el is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# miscbits-el is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with miscbits-el; see the file COPYING.  Failing that, go to
# <http://www.gnu.org/licenses/>.

use strict;
use warnings;
use File::Basename 'dirname';
use File::Spec;
use HTML::Parser;
use POSIX;
use Perl6::Slurp;
use IPC::Run;
use Iterator::Simple qw(igrep);
use Iterator::Simple::Locate;

use lib::abs '.';
use MyStuff;
use MyUniqByInode;

my $verbose = 0;

my $it = Iterator::Simple::Locate->new (globs => [
                                                  '/usr/share/doc/*.html',
                                                  '/usr/share/doc/*.htm',
                                                  # '/z/usr/share/doc/*.html',
                                                  # '/z/usr/share/doc/*.htm',
                                                 ]);
# {
#   my $uniq = MyUniqByInode->new;
#   $it = igrep { $uniq->uniq($_) } $it;
# }

{
  # look for <link rel=""> missing targets
  my ($filename, $dirname);
  my %seen;

  my $start_handler = sub {
    my ($tagname, $attr, $linenum) = @_;

    my $target = $attr->{'href'} // return;
    $target =~ s/^\s+//;
    $target =~ s/\s+$//;
    if ($target =~ /[a-z]:/) {
      $target =~ s/^file:// or return;
    }
    $target =~ s/#.*$//;
    return if $target eq '';  # fragment only
    my $target_fullname = File::Spec->rel2abs($target,$dirname);
    if (! -f $target_fullname) {
      my $rel = lc($attr->{'rel'}//'');
      print "$filename:$linenum: rel=$rel href=$attr->{'href'} (at $target_fullname)\n";
    }
  };
  my $parser = HTML::Parser->new
    (api_version => 3,
     start_h => [ $start_handler, 'tagname,attr,line' ]);
  $parser->report_tags('link');

  my $count = 0;
  while ($filename = $it->next) {
    if ($verbose) { print "look at $filename\n"; }
    $dirname = dirname($filename);
    $parser->parse_file($filename);
  }
  print "total links $count\n";
}
{
  # look for <link rel=""> types
  my $filename;
  my %seen;

  my $start_handler = sub {
    my ($tagname, $attr, $linenum) = @_;

    my $rel = lc($attr->{'rev'}//'');
    # if ($rel eq '' || ! $seen{$rel}++) {
    if (! $seen{$rel}++) {
      print "$filename:$linenum: rel=$rel\n";
    }
  };
  my $parser = HTML::Parser->new
    (api_version => 3,
     start_h => [ $start_handler, 'tagname,attr,line' ]);
  $parser->report_tags('link');

  my $count = 0;
  while ($filename = $it->next) {
    if ($verbose) { print "look at $filename\n"; }
    my $dirname = dirname($filename);

    $parser->parse_file($filename);
  }
  print "total links $count\n";
}
{
  # look for width=1,height=1 bugs
  my $filename;

  my $start_handler = sub {
    my ($tagname, $attr, $linenum) = @_;

    if (($attr->{'width'}//'') eq 1 && ($attr->{'height'}//'') eq 1) {
      print "$filename:$linenum: width=1 height=1\n";
    }
  };
  my $parser = HTML::Parser->new
    (api_version => 3,
     start_h => [ $start_handler, 'tagname,attr,line' ]);
  $parser->report_tags('img');

  my $count = 0;
  while ($filename = $it->next) {
    if ($verbose) { print "look at $filename\n"; }
    my $dirname = dirname($filename);

    $parser->parse_file($filename);
  }
  print "total links $count\n";
}

{
  my $count = 0;
  while (my $filename = $it->next) {
    if ($verbose) { print "look at $filename\n"; }
    my $dirname = dirname($filename);

    my $str;
    if (! eval { $str = Perl6::Slurp::slurp($filename); 1 }) {
      unless ($! == ENOENT) {
        print "Cannot read $filename: $!\n";
      }
      next;
    }

    # src=""
    #               1          2       3        4
    while ($str =~ /(src|href)=(\s*['"]([^'"]*)|([^ >]*))/g) {
      my $type = $1;
      next unless $type eq 'src';  # <img> only
      my $pos = $-[0];
      my $target = $3 // $4;

      if ($target eq '') {
        my ($line, $col) = MyStuff::pos_to_line_and_column ($str, $pos);
        print "$filename:$line:$col: empty link target\n";
        next;
      }
      if ($target =~ /\n/) {
        my ($line, $col) = MyStuff::pos_to_line_and_column ($str, $pos);
        print "$filename:$line:$col: newline in link $target\n";
        next;
      }
      $target =~ s{^\./}{}; # drop leading current directory
      $target =~ s{#.*}{};  # drop fragment

      if ($target =~ /:/) {
        # http: etc schema
        if ($type eq 'src') {
          my ($line, $col) = MyStuff::pos_to_line_and_column ($str, $pos);
          print "$filename:$line:$col: image with schema $target\n";
        }
      }
      next;

      next unless $target =~ /\Q..\E/;  # higher directory
      # next if $target =~ m{/};    # sub-directory

      next if $target =~ m{\.js$};
      next if $target eq '';
      $count++;

      my $target_fullname = File::Spec->catfile($dirname, $target);
      next if -e $target_fullname;  # exists, ok

      my ($line, $col) = MyStuff::pos_to_line_and_column ($str, $pos);
      print "$filename:$line:$col: missing $target_fullname\n";
      my $l = File::Locate::Iterator->new (regexp => qr{/\Q$target\E$});
      my $count = 0;
      while (my ($found) = $l->next) {
        print "  maybe: $found\n";
        last if ++$count >= 3;
      }
    }
  }
  print "total links $count\n";
}
exit 0;
